/* ------------------------------------------------------------------------- */
/* Version history:                                                          */
/* 01.12.2005 : v1.0 public release                                          */
/* ------------------------------------------------------------------------- */

#include <windows.h>
#include <memory.h>
#include <commctrl.h>
#include "resource.h"
#include "lzxdec.h"

#ifdef _DEBUG
# include <stdlib.h>
# include <stdio.h>
#else /* some size optimization :-) */
# pragma comment(linker, "/NODEFAULTLIB /FILEALIGN:0x200 /ENTRY:xWinMain")
# pragma comment(linker, "/MERGE:.rdata=.text /MERGE:.data=.text")
# pragma comment(linker, "/MERGE:.data1=.text /IGNORE:4078 /SECTION:.text,ER")
#endif

/* ------------------------------------------------------------------------- */
#include <pshpack1.h>
/* ------------------------------------------------------------------------- */

#define MAX_FILENAME    (512)          /* max openfilename path     */
#define MIN_FILE_ALIGN  (512)          /* minimum section alignment */

/* error codes for unpacker thread */
#define ERR_OK             (0)
#define ERR_NO_FILE        (1)
#define ERR_FILE_OPEN      (2)
#define ERR_NO_MEM         (3)
#define ERR_CANT_MAP       (4)
#define ERR_FILE_CREATE    (5)
#define ERR_FILE_WRITE     (6)
#define ERR_COPY_OVL       (7)
#define ERR_CORRUPT        (8)
#define ERR_UNKNOWN        (9)
#define ERR_SKIPPED        (10)
#define ERR_UNPACK         (11)
#define ERR_FAILED         (12)

/* aspack version info */
typedef struct _ASPACK_INFO {
  DWORD  Crc1Ofs;                      /* first checksum offset           */
  DWORD  Crc1Len;                      /* first checksum length           */
  DWORD  Crc1Val;                      /* first checksum value            */
  DWORD  Crc2Ofs;                      /* second checksum offset          */
  DWORD  Crc2Len;                      /* second checksum length          */
  DWORD  Crc2Val;                      /* second checksum value           */
  CHAR*  VerName;                      /* name of this version            */
  DWORD  ObjTbl;                       /* object table offset             */
  DWORD  FlgE8E9;                      /* e8/e9 filter flag offset        */
  DWORD  ModE8E9;                      /* e8/e9 filter mode offset        */
  DWORD  CmpE8E9;                      /* e8/e9 filter mark offset        */
  DWORD  RelTbl;                       /* offset of relocation table rva  */
  DWORD  ImpTbl;                       /* offset of import table rva      */
  DWORD  OepOfs;                       /* offset of entry point rva       */
} ASPACK_INFO, *PASPACK_INFO;

#define  ASP_VER_UNKNOWN  (-1)         /* unknown version id              */

/* versions data array */
static const ASPACK_INFO AspInfos[] = {
  { /* aspack v2.12 */
    0x39e,                   /* first checksum offset           */
    0x21,                    /* first checksum length           */
    0x98604df5,              /* first checksum value            */
    0x9d,                    /* second checksum offset          */
    0x62,                    /* second checksum length          */
    0xa82446ae,              /* second checksum value           */
    "v2.12",                 /* name of this version            */
    0x57b,                   /* object table offset             */
    0xfe,                    /* e8/e9 filter flag offset        */
    0x144,                   /* e8/e9 filter mode offset        */
    0x147,                   /* e8/e9 filter mark offset        */
    0x54b,                   /* offset of relocation table rva  */
    0x278,                   /* offset of import table rva      */
    0x39a                    /* offset of entry point rva       */
  }
};

/* aspack object entry */
typedef struct _ASP_OBJ {
  DWORD  ObjRVA;                       /* section rva           */
  DWORD  ObjSize;                      /* section unpacked size */
} ASP_OBJ, *PASP_OBJ;

/* work context */
typedef struct _UNASP_CTX {
  HINSTANCE  hInst;                    /* self instance                   */
  HWND  hDlg;                          /* main dialog handle              */
  HWND  hInfo;                         /* info edit-box handle            */
  INT   Vers;                          /* aspack version                  */
  BYTE* Image;                         /* working image                   */
  BYTE* Entry;                         /* entry-point ptr                 */
  UINT  OvlOffs;                       /* file overlay offset             */
  UINT  OvlSize;                       /* file overlay size               */
  UINT  WrapRVA;                       /* wrapper section rva             */
  UINT  WrapSize;                      /* wrapper section size            */
  CHAR* InpName;                       /* input file name (silent mode)   */
  CHAR* OutName;                       /* output file name (silent mode)  */
  HANDLE hFile;                        /* original file handle            */
  ASPACK_INFO Info;                    /* this version info               */
  IMAGE_NT_HEADERS* PE;                /* working pe header ptr           */
  IMAGE_SECTION_HEADER* Obj;           /* working section headers ptr     */
  CHAR  FName[MAX_FILENAME+4];         /* original file name              */
} UNASP_CTX, *PUNASP_CTX;

/* some string constants */
static const CHAR ERROR_TIT[] = "ASPack Unpacking Error:";
static const CHAR ABOUT_TIT[] = "About this unpacker...";
static const CHAR DLG_TITLE[] = "Static ASPack Unpacker v1.00";
static const CHAR UTHRD_ERR[] = "Can't create decompressor thread";
static const CHAR OFN_FILTR[] = "Any Win32 PE Files (*.exe,*.scr,*.dll,*.ocx)\
\0*.exe;*.scr;*.dll;*.ocx\0Any File (*.*)\0*.*\0\0";
static const CHAR ABOUT_TXT[] = "ASPack Unpacker v1.00 by Dr.Golova\r\n\
Supported ASPack v2.12\r\nmailto: dr_golova@mail.ru";

/* useful macro */
#define ALIGN(X,Y)    (((X)+(Y)-1)&(~((Y)-1)))
#define MIN(A,B)      ((A)>(B)?(B):(A))
#define MAX(A,B)      ((A)>(B)?(A):(B))
#define ROL(X,N)      (((X)<<(N))|((X)>>(8*sizeof(X)-(N))))

/* ------------------------------------------------------------------------- */
#include <poppack.h>
/* ------------------------------------------------------------------------- */

/* small block allocator */
static BYTE* SmallAlloc(UINT uBlkSize)
{
#ifndef _DEBUG
  HANDLE hHeap;
  LPVOID pMem;
  
  hHeap = GetProcessHeap();
  pMem  = HeapAlloc(hHeap, HEAP_ZERO_MEMORY, uBlkSize);
  return(BYTE*)(pMem);
#else
  return(BYTE*)malloc(uBlkSize);
#endif
}

static VOID SmallFree(LPVOID pMem)
{
#ifndef _DEBUG
  HANDLE hHeap;

  hHeap = GetProcessHeap();
  HeapFree(hHeap, 0, pMem);
#else
  free(pMem);
#endif
}

/* ------------------------------------------------------------------------- */

/* big block allocator */
static BYTE* BigAlloc(UINT uBlkSize)
{
#ifndef _DEBUG
  LPVOID pMem;

  pMem = VirtualAlloc(NULL, uBlkSize, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
  return(BYTE*)(pMem);
#else
  return(BYTE*)calloc(1, uBlkSize);
#endif
}

static VOID BigFree(LPVOID pMem)
{
#ifndef _DEBUG
  VirtualFree(pMem, 0, MEM_DECOMMIT);
#else
  free(pMem);
#endif
}

/* ------------------------------------------------------------------------- */

static VOID dg_memset(VOID* pMem, INT Val, UINT uLen)
{
  while( uLen-- )
  {
    *(CHAR*)(pMem) = (CHAR)(Val);
    pMem = (CHAR*)(pMem) + 1;
  }
}

static VOID dg_memcpy(VOID* pDst, CONST VOID* pSrc, UINT uLen)
{
  while( uLen-- )
  {
    *(CHAR*)(pDst) = *(CHAR*)(pSrc);
    pSrc = (CHAR*)(pSrc) + 1;
    pDst = (CHAR*)(pDst) + 1;
  }
}

/* ------------------------------------------------------------------------- */

/* free all used resources */
static VOID ReleaseContext(UNASP_CTX* Ctx)
{
  if ( NULL != Ctx->Image )
    BigFree(Ctx->Image), Ctx->Image = (NULL);
  if ( NULL != Ctx->hFile )
    CloseHandle(Ctx->hFile), Ctx->hFile = (NULL);
  
  Ctx->OvlOffs = 0;
  Ctx->OvlSize = 0;
  Ctx->Entry = NULL;
}

/* ##### *** BOYER-MOORE FAST EXACT PATTERN SEARCH *** ##################### */

/* max pattern length */
#define MAX_BMS_PATTERN  256

/* boyer-moor context */
typedef struct _BMS_CTX {
  BYTE   Table[MAX_BMS_PATTERN];
  UINT   PatLen;
  UINT   Skip[256];
} BMS_CTX, *PBMS_CTX;

/* ------------------------------------------------------------------------- */

/* setup boyer-moor context */
static BOOL BMS_SetPattern(BMS_CTX* BMS, CONST BYTE* Pattern, UINT Length)
{
  UINT I;
  
  /* check params */
  if ( NULL == BMS || Length > MAX_BMS_PATTERN )
    return(FALSE);
  
  BMS->PatLen = Length;
  dg_memcpy(BMS->Table, Pattern, Length);

  /* build skip table */
  for ( I = 0; I < 256; I++ )
    BMS->Skip[I] = Length;
  for ( I = 1; I <= Length; I++ )
    BMS->Skip[Pattern[I-1]] = (Length - I);
  
  return(TRUE);
}

/* ------------------------------------------------------------------------- */

/* scan code for pattern, return ptr or null */
static BYTE* BMS_Find(BMS_CTX* BMS, CONST BYTE* Code, UINT CodeLen)
{
  BYTE* Table;
  UINT* Skip;
  UINT  I, J;
  UINT  PatLen;
  INT   A, B;
  
  I = J = PatLen = BMS->PatLen;
  Table = BMS->Table;
  Skip  = BMS->Skip;
  
  do
  {
    if ( Code[I-1] == Table[J-1] )
    {
      I--; J--;
    }
    else
    {
      A = PatLen - J + 1;
      B = Skip[Code[I-1]];
      I += MAX(A, B);
      J = PatLen;
    }
  }
  while( J >= 1 && I <= CodeLen );
  
  return (BYTE*)((I >= CodeLen) ? (NULL) : (&Code[I]));
}

/* ##### *** WIN32 PE IMAGE RELATED FUNCTIONS *** ########################## */

/* map input file image */
static INT MapPackedFile(UNASP_CTX* Ctx)
{
  IMAGE_DOS_HEADER* DOS;
  IMAGE_NT_HEADERS* PE;
  IMAGE_SECTION_HEADER* Obj;
  HANDLE hFile, hMap;
  UINT   ISize, I, FMax, PECnt;
  UINT   OEP, RA, VA;
  UINT   FLen;
  BYTE*  pMap;

  ReleaseContext(Ctx);
  if ( INVALID_HANDLE_VALUE == (hFile = CreateFile(Ctx->FName, GENERIC_READ,
    0, NULL, OPEN_EXISTING, 0, NULL)) )
  {
    return(ERR_FILE_OPEN);
  }
  if ( NULL == (hMap = CreateFileMapping(hFile,NULL,PAGE_READONLY,0,0,NULL)) )
  {
    CloseHandle(hFile);
    return(ERR_CANT_MAP);
  }

  if ( NULL == (pMap = (BYTE*)MapViewOfFile(hMap,FILE_MAP_READ,0,0,0)) )
  {
    CloseHandle(hMap);
    CloseHandle(hFile);
    return(ERR_CANT_MAP);
  }
  
  FLen = GetFileSize(hFile, NULL);
  DOS  = (IMAGE_DOS_HEADER*)(pMap);
  if ( FLen < sizeof(IMAGE_DOS_HEADER) || (UINT)(DOS->e_lfanew + 4) > FLen ||
       DOS->e_magic != IMAGE_DOS_SIGNATURE || (UINT)DOS->e_lfanew > FLen )
  {
    UnmapViewOfFile(pMap);
    CloseHandle(hMap);
    CloseHandle(hFile);
    return(ERR_CANT_MAP); /* not mz-image */
  }

  PE = (IMAGE_NT_HEADERS*)(&pMap[DOS->e_lfanew]);
  if ( PE->Signature != IMAGE_NT_SIGNATURE )
  {
    UnmapViewOfFile(pMap);
    CloseHandle(hMap);
    CloseHandle(hFile);
    return(ERR_CANT_MAP); /* not pe-image */
  }

  ISize = PE->OptionalHeader.SizeOfImage;
  if ( NULL == (Ctx->Image = BigAlloc(ISize + 0x10000)) )
  {
    UnmapViewOfFile(pMap);
    CloseHandle(hMap);
    CloseHandle(hFile);
    return(ERR_NO_MEM);
  }

  /* copy image headers */
  PECnt = PE->FileHeader.SizeOfOptionalHeader + sizeof(PE->FileHeader) + 4;
  dg_memcpy(Ctx->Image, pMap, DOS->e_lfanew + PECnt +
    (PE->FileHeader.NumberOfSections * sizeof(IMAGE_SECTION_HEADER)));
  PE  = Ctx->PE  = (IMAGE_NT_HEADERS*)(&Ctx->Image[DOS->e_lfanew]);
  Obj = Ctx->Obj = (IMAGE_SECTION_HEADER*)((BYTE*)(Ctx->PE) + PECnt);

  /* realign and find wrapper */
  Ctx->WrapRVA = Ctx->WrapSize = 0;
  OEP = PE->OptionalHeader.AddressOfEntryPoint;
  RA = PE->OptionalHeader.FileAlignment;
  VA = PE->OptionalHeader.SectionAlignment;
  for ( FMax = 0, I = 0; I < PE->FileHeader.NumberOfSections; I++ )
  {
    /* check for oep object */
    if ( OEP >= Obj[I].VirtualAddress &&
         OEP <  Obj[I].VirtualAddress + Obj[I].Misc.VirtualSize )
    {
      Ctx->WrapRVA  = Obj[I].VirtualAddress;
      Ctx->WrapSize = Obj[I].Misc.VirtualSize;
    }

    /* copy section to context image */
    dg_memcpy(&Ctx->Image[Obj[I].VirtualAddress],&pMap[Obj[I].
      PointerToRawData], Obj[I].SizeOfRawData);

    /* update section header */
    if ( FMax < (Obj[I].PointerToRawData + Obj[I].SizeOfRawData) )
      FMax = (Obj[I].PointerToRawData + Obj[I].SizeOfRawData);
    Obj[I].SizeOfRawData = Obj[I].Misc.VirtualSize =
      ALIGN(Obj[I].Misc.VirtualSize, VA);
    Obj[I].PointerToRawData = Obj[I].VirtualAddress;
  }

  UnmapViewOfFile(pMap);
  CloseHandle(hMap);

  /* check for overlay */
  if ( 0 != (Ctx->OvlSize = GetFileSize(hFile, NULL) - FMax) )
  {
    Ctx->OvlOffs = FMax;
    Ctx->hFile = hFile;
  }
  else
  {
    CloseHandle(hFile);
    Ctx->hFile = (NULL);
  }

  return(ERR_OK); /* all ok */
}

/* ------------------------------------------------------------------------- */

/* save map to file, realign sections if needed */
static INT SaveMappedFile(UNASP_CTX* Ctx, CHAR* FName)
{
  IMAGE_DATA_DIRECTORY* DDir;
  IMAGE_NT_HEADERS* PE;
  IMAGE_SECTION_HEADER* Obj;
  HANDLE hFile;
  DWORD  Wrt, Rdt, BSize, Cnt;
  DWORD  Offs, Size, HSize;
  BYTE*  BPtr, *SPtr;
  BYTE*  Buff;
  INT    I;

  if ( INVALID_HANDLE_VALUE == (hFile = CreateFile(FName, GENERIC_WRITE,
    0, NULL, CREATE_ALWAYS, 0, NULL)) )
  {
    return(ERR_FILE_CREATE);
  }

  SetFilePointer(hFile, 0, NULL, FILE_BEGIN);

  PE = Ctx->PE;
  PE->OptionalHeader.SizeOfImage = ALIGN(PE->OptionalHeader.SizeOfImage,
    PE->OptionalHeader.SectionAlignment);

  DDir = &PE->OptionalHeader.DataDirectory[0];
  DDir[IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT].Size = 0;
  DDir[IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT].VirtualAddress = 0;
  DDir[IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT].Size = 0;
  DDir[IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT].VirtualAddress = 0;

  Obj = Ctx->Obj;

  /* realign sections */
  PE->OptionalHeader.FileAlignment = MIN(MIN_FILE_ALIGN,
    PE->OptionalHeader.FileAlignment);

  /* calc header size/first section offset */
  HSize = (DWORD)((BYTE*)(Ctx->Obj) - Ctx->Image + 
    (sizeof(IMAGE_SECTION_HEADER) * PE->FileHeader.NumberOfSections));
  Offs  = ALIGN(HSize, PE->OptionalHeader.FileAlignment);

  /* write sections */
  for ( I = 0; I < PE->FileHeader.NumberOfSections; I++ )
  {
    SPtr = (Ctx->Image + Obj[I].VirtualAddress);
    BPtr = (SPtr + Obj[I].Misc.VirtualSize - 1);
    while( (BPtr > SPtr) && *BPtr == 0 ) BPtr--;
    if ( BPtr < SPtr ) BPtr = SPtr;
    if ( BPtr > SPtr ) BPtr++;
    SetFilePointer(hFile, Offs, NULL, FILE_BEGIN);
    Size = ALIGN(BPtr - SPtr, PE->OptionalHeader.FileAlignment);
    if ( FALSE == WriteFile(hFile, SPtr, Size, &Wrt, NULL) || Wrt != Size )
    {
      CloseHandle(hFile);
      return(ERR_FILE_WRITE);
    }
    Obj[I].PointerToRawData = Offs;
    Obj[I].SizeOfRawData    = Size;
    Offs += Size;
  }

  /* update header */
  SetFilePointer(hFile, 0, NULL, FILE_BEGIN);
  if ( FALSE == WriteFile(hFile,Ctx->Image,HSize,&Wrt,NULL) || Wrt != HSize )
  {
    CloseHandle(hFile);
    return(ERR_FILE_WRITE);
  }

  if ( Ctx->OvlOffs && Ctx->OvlSize )
  {
    /* copy overlay */
    SetFilePointer(hFile, 0, NULL, FILE_END);
    SetFilePointer(Ctx->hFile, Ctx->OvlOffs, NULL, FILE_BEGIN);
    Size = Ctx->OvlSize;
    BSize = MIN(Size, 0x100000);
    if ( NULL == (Buff = BigAlloc(BSize)) )
    {
      CloseHandle(hFile);
      return(ERR_NO_MEM);
    }
    while( Size )
    {
      Cnt = MIN(Size, BSize);
      if ( FALSE == ReadFile(Ctx->hFile,Buff,Cnt,&Rdt,NULL) || Rdt != Cnt ||
           FALSE == WriteFile(hFile,Buff,Cnt,&Wrt,NULL) || Wrt != Cnt )
      {
        BigFree(Buff);
        CloseHandle(hFile);
        return(ERR_COPY_OVL);
      }
      Size -= Cnt;
    }
    BigFree(Buff);
  }

  /* all done */
  CloseHandle(hFile);
  return(ERR_OK);
}

/* ##### *** VERSION CHECKER *** ########################################### */

static DWORD CalcCrc32(VOID* pMem, UINT uLen, DWORD dwSeed)
{
  UINT I, C;

  while( uLen-- )
  {
    C = *(BYTE*)(pMem);
    pMem = (BYTE*)(pMem) + 1;

    for( I = 0; I < 8; I++ )
    {
      if ( (dwSeed ^ C) & 1 )
        dwSeed = (dwSeed >> 1) ^ 0xEDB88320;
      else
        dwSeed = (dwSeed >> 1);

      C >>= 1;
    }
  }

  return(dwSeed);
}

/* ------------------------------------------------------------------------- */

static INT GetASPVersion(LPVOID pEntry, UINT uMaxOfs)
{
  ASPACK_INFO* Info;
  DWORD Crc32;
  UINT  I;

  for ( I = 0; I < sizeof(AspInfos)/sizeof(AspInfos[0]); I++ )
  {
    /* check first crc info */
    Info = (ASPACK_INFO*)(&AspInfos[I]);
    if ( Info->Crc1Ofs >= uMaxOfs || Info->Crc1Len >= uMaxOfs )
      continue; /* overrun */
    if ( (Info->Crc1Ofs + Info->Crc1Len) > uMaxOfs )
      continue; /* overrun */

    /* compare first checksums */
    Crc32 = CalcCrc32((BYTE*)(pEntry)+Info->Crc1Ofs,Info->Crc1Len,0xffffffff);
    if ( Crc32 != Info->Crc1Val )
      continue; /* not same */

    /* check second crc info */
    if ( Info->Crc2Ofs >= uMaxOfs || Info->Crc2Len >= uMaxOfs )
      continue; /* overrun */
    if ( (Info->Crc2Ofs + Info->Crc2Len) > uMaxOfs )
      continue; /* overrun */
    
    /* compare second checksums */
    Crc32 = CalcCrc32((BYTE*)(pEntry)+Info->Crc2Ofs,Info->Crc2Len,0xffffffff);
    if ( Crc32 != Info->Crc2Val )
      continue; /* not same */

    return(I); /* all ok */
  }

  return(ASP_VER_UNKNOWN); /* not founded */
}

/* ##### *** STATIC UNPACKERS/DECRYPTORS *** ############################### */

static INT UnpackSections(UNASP_CTX* Ctx)
{
  LZX_CONTEXT* LZX = NULL;
  ASP_OBJ* Obj;
  BYTE* UMem = NULL;
  UINT  ULen = 0;
  BYTE* PMem;
  INT   ILen;

  /* check table rva */
  if ( !Ctx->Info.ObjTbl )
    return(ERR_SKIPPED);

  /* alloc lzx decoder context */
  if ( NULL == (LZX = (LZX_CONTEXT*)SmallAlloc(sizeof(LZX_CONTEXT))) )
    return(ERR_NO_MEM);

  /* process sections */
  Obj = (ASP_OBJ*)(Ctx->Entry + Ctx->Info.ObjTbl);
  while( Obj->ObjRVA != 0 )
  {
    /* check output buffer size */
    if ( UMem != NULL && ULen < Obj->ObjSize )
      BigFree(UMem), UMem = NULL;

    /* alloc buffer */
    if ( NULL == UMem )
    {
      ULen = Obj->ObjSize;
      if ( NULL == (UMem = BigAlloc(ULen + 300)) )
      {
        SmallFree(LZX);
        return(ERR_NO_MEM);
      }
    }

    /* unpack block */
    PMem = &Ctx->Image[Obj->ObjRVA];
    ILen = DecodeLZX(LZX, PMem, UMem, Obj->ObjSize, Obj->ObjSize);
    if ( ILen < 0 || (UINT)(ILen) < Obj->ObjSize )
    {
      BigFree(UMem);
      SmallFree(LZX);
      return(ERR_UNPACK);
    }

    dg_memcpy(PMem, UMem, Obj->ObjSize);
    Obj++; /* to next obj */
  }

  SmallFree(LZX);
  return(ERR_OK);
}

/* ------------------------------------------------------------------------- */

/* decode e8/e9 optimization filter */
static INT DecodeE8E9(UNASP_CTX* Ctx)
{
  ASP_OBJ* Obj;
  DWORD DVal;
  BYTE* MPtr;
  UINT  Size;
  UINT  Offs;
  INT   Mode;
  BYTE  BCmp;
  BYTE  B0;

  /* check filter ofs */
  if ( !Ctx->Info.FlgE8E9 )
    return(ERR_SKIPPED);
  if ( Ctx->Entry[Ctx->Info.FlgE8E9] != 0 )
    return(ERR_SKIPPED);

  /* check first obj rva */
  if ( !Ctx->Info.ObjTbl )
    return(ERR_SKIPPED);
  Obj = (ASP_OBJ*)(&Ctx->Entry[Ctx->Info.ObjTbl]);
  if ( !Obj->ObjRVA || Obj->ObjSize < 6 )
    return(ERR_SKIPPED);

  Offs = 0; /* decode */
  Size = (Obj->ObjSize - 6);
  MPtr = (Ctx->Image + Obj->ObjRVA);
  BCmp = (Ctx->Entry[Ctx->Info.CmpE8E9]);
  Mode = (Ctx->Entry[Ctx->Info.ModE8E9] == 0) ? (0) : (1);
  while( Size )
  {
    B0 = MPtr[Offs];
    if ( B0 != 0xE8 && B0 != 0xE9 )
    {
      Size--; Offs++;
      continue;
    }

    DVal = *(DWORD*)(&MPtr[Offs+1]);
    if ( Mode == 0 )
    {
      if ( (DVal & 0xff) != BCmp )
      {
        Size--; Offs++;
        continue;
      }
      DVal &= 0xffffff00;
      DVal = ROL(DVal, 24);
    }

    *(DWORD*)(&MPtr[Offs+1]) = DVal - Offs;
    Offs += 5; Size -= MIN(Size, 5);
  }

  return(ERR_OK);
}

/* ------------------------------------------------------------------------- */

/* reconstruct fixup table */
static INT RebuildRelocs(UNASP_CTX* Ctx)
{
  DWORD* RPtr;
  DWORD  RVA;
  UINT   Size;

  /* flush relocations */
  Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].
    VirtualAddress = 0;
  Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].
    Size = 0;

  /* check offs */
  if ( !Ctx->Info.RelTbl )
    return(ERR_SKIPPED);
  if ( 0 == (RVA = *(DWORD*)(&Ctx->Entry[Ctx->Info.RelTbl])) )
    return(ERR_SKIPPED);

  Size = 0; /* calc size */
  RPtr = (DWORD*)(&Ctx->Image[RVA]);
  while( RPtr[0] != 0 && RPtr[1] != 0 )
  {
    Size += RPtr[1];
    RPtr = (DWORD*)((BYTE*)(RPtr) + RPtr[1]);
  }
  
  /* check dummy relocs */
  if ( Size == 0 ) RVA = 0;

  /* update header */
  Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].
    VirtualAddress = RVA;
  Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].
    Size = Size;

  return(ERR_OK);
}

/* ------------------------------------------------------------------------- */

/* reconstruct import table */
static INT RebuildImport(UNASP_CTX* Ctx)
{
  IMAGE_IMPORT_DESCRIPTOR* IDir;
  DWORD RVA;
  DWORD Size;

  /* flush import table */
  Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].
    VirtualAddress = 0;
  Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].
    Size = 0;

  /* check offs */
  if ( !Ctx->Info.ImpTbl )
    return(ERR_SKIPPED);
  if ( 0 == (RVA = *(DWORD*)(&Ctx->Entry[Ctx->Info.ImpTbl])) )
    return(ERR_SKIPPED);
  
  /* process table */
  Size = sizeof(IMAGE_IMPORT_DESCRIPTOR);
  IDir = (IMAGE_IMPORT_DESCRIPTOR*)(&Ctx->Image[RVA]);
  while( 0 != IDir->Name )
  {
    Size += sizeof(IMAGE_IMPORT_DESCRIPTOR);
    IDir = (IMAGE_IMPORT_DESCRIPTOR*)((BYTE*)(IDir) +
      sizeof(IMAGE_IMPORT_DESCRIPTOR));
  }
  
  /* update header */
  Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].
    VirtualAddress = RVA;
  Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].
    Size = Size;

  return(ERR_OK);
}

/* ------------------------------------------------------------------------- */

/* update header (oep at this time) */
static INT UpdateHeaders(UNASP_CTX* Ctx)
{
  DWORD RVA;

  /* flush oep */
  Ctx->PE->OptionalHeader.AddressOfEntryPoint = 0;

  /* check offs */
  if ( !Ctx->Info.OepOfs )
    return(ERR_SKIPPED);

  /* update header */
  RVA = *(DWORD*)(&Ctx->Entry[Ctx->Info.OepOfs]);
  Ctx->PE->OptionalHeader.AddressOfEntryPoint = RVA;

  return(ERR_OK);
}

/* ------------------------------------------------------------------------- */

/* move tls to original place */
static INT RebuildTLS(UNASP_CTX* Ctx)
{
  IMAGE_DATA_DIRECTORY* ImgDir;
  BMS_CTX* BMS;
  DWORD  TlsRVA;
  BYTE*  OrgTls;
  UINT   TlsLen;

  ImgDir = &Ctx->PE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_TLS];
  if ( !(TlsRVA = ImgDir->VirtualAddress) || !(TlsLen = ImgDir->Size) )
  {
    /* no tls in image */
    ImgDir->VirtualAddress = 0;
    ImgDir->Size = 0;
    return(ERR_OK);
  }

  /* alloc bms context */
  if ( NULL == (BMS = (BMS_CTX*)SmallAlloc(sizeof(BMS_CTX))) )
    return(ERR_NO_MEM);

  /* setup bms context */
  if ( !BMS_SetPattern(BMS, &Ctx->Image[TlsRVA], TlsLen) )
  {
    SmallFree(BMS);
    return(ERR_FAILED);
  }

  /* find original tls before original entry */
  OrgTls = BMS_Find(BMS, Ctx->Image, Ctx->WrapRVA);
  SmallFree(BMS);

  /* check result */
  if ( NULL == OrgTls )
    return(ERR_FAILED);

  /* update header */
  ImgDir->VirtualAddress = (DWORD)(OrgTls - Ctx->Image);
  return(ERR_OK);
}

/* ------------------------------------------------------------------------- */

#if 0
/* rebuild whole image */
static INT RebuildImage(UNASP_CTX* Ctx)
{
  IMAGE_SECTION_HEADER* Obj;
  DWORD  OEP;
  INT    Res;

  /* reconstruct image */
  if ( ERR_OK != (Res = RebuildTLS(Ctx)) )
    return(Res);
  if ( ERR_OK != (Res = RebuildRsrc(Ctx)) )
    return(Res);
  if ( ERR_OK != (Res = DeleteWrapperObj(Ctx)) )
    return(Res);

  /* mark oep section as code section */
  OEP = Ctx->PE->OptionalHeader.AddressOfEntryPoint;
  if ( NULL == (Obj = FindObjByRVA(Ctx->PE, OEP)) )
    return(ERR_FAILED); /* corrupt */
  Obj->Characteristics |= (IMAGE_SCN_CNT_CODE|IMAGE_SCN_MEM_EXECUTE);

  return(ERR_OK);
}
#endif

/* ##### *** GUI INFO BOX RELATED FUNCTIONS *** ############################ */

/* add text to information box */
static VOID AddInfoText(HWND hEdit, CONST CHAR* Text, INT CrLf)
{
  CHAR MsgBuff[2048];
  INT  I;

  if ( NULL == hEdit )
    return; /* silent */
  
  MsgBuff[0] = '\0';
  SendMessage(hEdit, WM_GETTEXT, sizeof(MsgBuff), (LPARAM)(&MsgBuff[0]));
  lstrcat(&MsgBuff[0], Text);
  if ( CrLf > 0 )  
  {
    for ( I = 0; I < CrLf; I++ )
      lstrcat(&MsgBuff[0], "\r\n");
  }
  SendMessage(hEdit, WM_SETTEXT, 0, (LPARAM)(&MsgBuff[0]));
}

/* ------------------------------------------------------------------------- */

/* parse and display posible error */
static INT AddInfoError(UNASP_CTX* Ctx, INT Err)
{
  CHAR* Message;
  
  if ( ERR_OK == Err )
  {
    Message = "Ok";
  }
  else if ( ERR_SKIPPED == Err )
  {
    Message = "Skipped";
    Err = (ERR_OK);
  }
  else
  {
    AddInfoText(Ctx->hInfo, "Failed", 1);
    switch(Err)
    {
    case ERR_FILE_OPEN:
      Message = "Can't open input file";
      break;
    case ERR_NO_MEM:
      Message = "Not enought memory";
      break;
    case ERR_CANT_MAP:
      Message = "Can't map image. Corrupt?";
      break;
    case ERR_FILE_CREATE:
      Message = "Can't create file. Write protected?";
      break;
    case ERR_FILE_WRITE:
      Message = "Can't write block to file. Corrupt?";
      break;
    case ERR_COPY_OVL:
      Message = "Can't copy overlay. Corrupt?";
      break;
    case ERR_CORRUPT:
      Message = "Corrupted data";
      break;
    case ERR_UNPACK:
      Message = "Posible corrupted data";
      break;
    case ERR_UNKNOWN:
      Message = "Unknown internal error";
    default:
      return(Err);
    }
  }
  
  AddInfoText(Ctx->hInfo, Message, 1);
  return(Err);
}

/* ##### *** MAIN DECOMPRESSOR THREAD *** ################################## */

/* main unpacker thread function */
static DWORD CALLBACK UThrdProc(LPVOID lpParam)
{
  CHAR FName[MAX_FILENAME+8];
  OPENFILENAME SFN;
  UNASP_CTX* Ctx;
  BYTE* OEP;
  UINT  Max;
  INT   Err;
  INT   Len;

  FName[0] = '\0';
  Ctx = (UNASP_CTX*)(lpParam);
  if ( Ctx->hInfo) SetWindowText(Ctx->hInfo, FName);
  AddInfoText(Ctx->hInfo, ABOUT_TXT, 2);
  if ( Ctx->InpName != NULL )
  {
    Len = lstrlen(Ctx->InpName);
    Len = MIN(Len, MAX_FILENAME);
    dg_memcpy(Ctx->FName, Ctx->InpName, Len);
    Ctx->FName[Len] = '\0';
  }

  if ( Ctx->FName[0] == '\0' )
  {
    AddInfoText(Ctx->hInfo, "Select packed file first !!!", 1);
    return(0);
  }

  /* map packed image */
  AddInfoText(Ctx->hInfo, "Mapping input image... ", 0);
  Err = MapPackedFile(Ctx);
  if ( ERR_OK != AddInfoError(Ctx, Err) )
    return(0); /* error */

  /* check version */
  Ctx->Entry = OEP = &Ctx->Image[Ctx->PE->OptionalHeader.AddressOfEntryPoint];
  Max = &Ctx->Image[Ctx->PE->OptionalHeader.SizeOfImage] - OEP;
  AddInfoText(Ctx->hInfo, "Checking version... ", 0);
  Ctx->Vers = GetASPVersion(OEP, Max);
  if ( Ctx->Vers == ASP_VER_UNKNOWN )
  {
    AddInfoText(Ctx->hInfo, "Failed", 1);
    AddInfoText(Ctx->hInfo, "Unknown ASPack version or not packed", 1);
    AddInfoText(Ctx->hInfo, "May be hacked/perverted ASPack", 1);
    return(0); /* error */
  }

  dg_memcpy(&Ctx->Info, &AspInfos[Ctx->Vers], sizeof(AspInfos[0]));

  /* drop version info */
  if ( NULL == Ctx->Info.VerName )
    AddInfoText(Ctx->hInfo, "Ok", 1);
  else
  {
    AddInfoText(Ctx->hInfo, "Ok [", 0);
    AddInfoText(Ctx->hInfo, Ctx->Info.VerName, 0);
    AddInfoText(Ctx->hInfo, "]", 1);
  }

  /* unpack sections */
  AddInfoText(Ctx->hInfo, "Unpacking sections... ", 0);
  Err = UnpackSections(Ctx);
  if ( ERR_OK != AddInfoError(Ctx, Err) )
    return(0); /* error */

  /* decode e8/e9 filter */
  AddInfoText(Ctx->hInfo, "Decode e8/e9 filter... ", 0);
  Err = DecodeE8E9(Ctx);
  if ( ERR_OK != AddInfoError(Ctx, Err) )
    return(0); /* error */

  /* reconstruct relocations */
  AddInfoText(Ctx->hInfo, "Rebuild relocations... ", 0);
  Err = RebuildRelocs(Ctx);
  if ( ERR_OK != AddInfoError(Ctx, Err) )
    return(0); /* error */

  /* reconstruct import table */
  AddInfoText(Ctx->hInfo, "Rebuild import table... ", 0);
  Err = RebuildImport(Ctx);
  if ( ERR_OK != AddInfoError(Ctx, Err) )
    return(0); /* error */

  /* reconstruct tls */
  AddInfoText(Ctx->hInfo, "Rebuild tls table... ", 0);
  Err = RebuildTLS(Ctx);
  if ( ERR_OK != AddInfoError(Ctx, Err) )
    return(0); /* error */

  /* update entry point */
  AddInfoText(Ctx->hInfo, "Update headers... ", 0);
  Err = UpdateHeaders(Ctx);
  if ( ERR_OK != AddInfoError(Ctx, Err) )
    return(0); /* error */

  /* save result image */
  if ( Ctx->OutName == NULL )
  {
    lstrcpy(FName, "unpacked.exe");
    dg_memset(&SFN, 0, sizeof(OPENFILENAME));
    SFN.lStructSize = sizeof(OPENFILENAME);
    SFN.hwndOwner   = Ctx->hDlg;
    SFN.lpstrFilter = OFN_FILTR;
    SFN.lpstrTitle  = "Save unpacked file as...";
    SFN.hInstance   = Ctx->hInst;
    SFN.lpstrFile   = &FName[0];
    SFN.nMaxFile    = (MAX_FILENAME);
    SFN.nFilterIndex = 2;
    SFN.Flags = (OFN_LONGNAMES | OFN_EXPLORER | OFN_HIDEREADONLY |
      OFN_OVERWRITEPROMPT|OFN_NOVALIDATE|OFN_NODEREFERENCELINKS|
      OFN_EXTENSIONDIFFERENT);
    AddInfoText(Ctx->hInfo, "Saving result... ", 0);
    if ( FALSE != GetSaveFileName(&SFN) )
    {
      Err = SaveMappedFile(Ctx, FName);
      if ( ERR_OK != AddInfoError(Ctx, Err) )
        return(0); /* error */
    }
    else
    {
      AddInfoText(Ctx->hInfo, "Cancelled", 1);
    }
  }
  else
  {
    Len = lstrlen(Ctx->OutName);
    Len = MAX(Len, MAX_FILENAME);
    dg_memcpy(FName, Ctx->OutName, Len);
    FName[Len] = '\0';

    Err = SaveMappedFile(Ctx, FName);
    if ( ERR_OK != AddInfoError(Ctx, Err) )
      return(0); /* error */
  }

  AddInfoText(Ctx->hInfo, "\r\nAll Done.", 1);
  ReleaseContext(Ctx);
  return(1); /* all ok */
}

/* ##### *** OTHER GUI CRAP *** ############################################ */

/* main dialog function */
static BOOL CALLBACK DlgProc(HWND hDlg,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  OPENFILENAME OFN;
  UNASP_CTX* Ctx;
  DWORD ThrdID;
  HICON hIcon;
  HFONT hFont;
  
  switch( uMsg )
  {
  case WM_INITDIALOG:
    Ctx = (UNASP_CTX*)(lParam);
    Ctx->hDlg  = hDlg;
    Ctx->hInfo = GetDlgItem(hDlg, IDC_INFO);
    hIcon = LoadIcon(Ctx->hInst, MAKEINTRESOURCE(IDI_MAINICON));
    SendMessage(hDlg, WM_SETICON, ICON_BIG,   (LPARAM)(hIcon));
    SendMessage(hDlg, WM_SETICON, ICON_SMALL, (LPARAM)(hIcon));
    SendMessage(hDlg, WM_SETTEXT, 0, (LPARAM)(DLG_TITLE));
    SetWindowLong(hDlg, DWL_USER, (LONG)(lParam));
    hFont = CreateFont(8,0,0,0,400,0,0,0,DEFAULT_CHARSET,OUT_DEFAULT_PRECIS,
      CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, "Terminal");
    if ( NULL != hFont ) 
      SendDlgItemMessage(hDlg, IDC_INFO, WM_SETFONT, (WPARAM)(hFont), 0);
    break;
  case WM_COMMAND:
    switch( LOWORD(wParam) )
    {
    case IDC_EXIT:
      EndDialog(hDlg, 0);
      break;
    case IDC_FILEBUTTON:
      Ctx = (UNASP_CTX*)GetWindowLong(hDlg, DWL_USER);
      Ctx->FName[0] = '\0';
      dg_memset(&OFN, 0, sizeof(OPENFILENAME));
      OFN.lStructSize = sizeof(OPENFILENAME);
      OFN.hwndOwner   = hDlg;
      OFN.lpstrFilter = OFN_FILTR;
      OFN.lpstrTitle  = "Select file for processing...";
      OFN.hInstance   = Ctx->hInst;
      OFN.lpstrFile   = &Ctx->FName[0];
      OFN.nMaxFile    = (MAX_FILENAME);
      OFN.nFilterIndex = 2;
      OFN.Flags = (OFN_LONGNAMES | OFN_EXPLORER | OFN_FILEMUSTEXIST |
        OFN_PATHMUSTEXIST | OFN_HIDEREADONLY);
      if ( FALSE != GetOpenFileName(&OFN) )
      {
        SetDlgItemText(hDlg, IDC_FILEEDIT, &Ctx->FName[0]);
      }
      break;
    case IDC_ABOUT:
      MessageBox(hDlg, ABOUT_TXT, ABOUT_TIT, MB_OK|MB_ICONINFORMATION);
      break;
    case IDC_UNPACK:
      Ctx = (UNASP_CTX*)GetWindowLong(hDlg, DWL_USER);
      Ctx->FName[0] = '\0';
      GetDlgItemText(hDlg, IDC_FILEEDIT, &Ctx->FName[0], MAX_FILENAME);
      if ( NULL == CreateThread(NULL, 0, UThrdProc, Ctx, 0, &ThrdID) )
      {
        MessageBox(hDlg, UTHRD_ERR, ERROR_TIT, MB_OK|MB_ICONSTOP);
      }
      break;
    default:
      return(FALSE);
    }
    break;
  case WM_CLOSE:
    EndDialog(hDlg, 0);
    break;
  default:
    return(FALSE);
  }
  
  return(TRUE);
}

/* ------------------------------------------------------------------------- */

/* fatal errors handler */
static LONG LastExceptionHandler(EXCEPTION_POINTERS* EPtr)
{
  MessageBox(GetForegroundWindow(), "Unhandled exception, sorry...",
    ERROR_TIT, (MB_OK|MB_ICONSTOP));
  return(EXCEPTION_EXECUTE_HANDLER);
}

/* ------------------------------------------------------------------------- */

/* basic commandline parser */
static VOID ExtractCmdLineParams(CHAR* CmdLine, CHAR** ArgV, CHAR* StrPtr,
                                 UINT* NumArgs, UINT* NumChar)
{
  CHAR* CPtr;
  UINT  ArgNum;
  UINT  ChrNum;
  INT   Quoted;
  INT   DoCopy;
  UINT  Num;
  CHAR  CH;

  ArgNum = 1;
  ChrNum = 0;
  CPtr = CmdLine;
  if ( ArgV != NULL )
    *ArgV++ = StrPtr;

  if ( (CH = *CPtr) == '\"' )
  {
    for ( ; ; )
    {
      CPtr++; CH = *CPtr;
      if ( CH == '\"' || CH == '\0' )
        break;

      ChrNum++;
      if ( StrPtr != NULL )
        *StrPtr++ = CH;
    }

    ChrNum++;
    if ( StrPtr != NULL )
      *StrPtr++ = '\0';

    if ( CH == '\"' )
      CPtr++;
  }
  else
  {
    for ( ; ; )
    {
      ChrNum++;
      if ( StrPtr != NULL )
        *StrPtr++ = CH;
      
      CPtr++;
      if ( CH <= ' ' )
        break;

      CH = *CPtr;
    }

    if ( CH == '\0' )
    {
      CPtr--;
    }
    else
    {
      ChrNum++;
      if ( StrPtr != NULL )
        *StrPtr++ = '\0';
    }
  }

  Quoted = 0;

  for ( ; ; )
  {
    if ( CH = *CPtr )
    {
      while ( CH == ' ' || CH == '\t' )
      {
        CPtr++;
        CH = *CPtr;
      }
    }

    if ( CH == '\0' )
      break; /* done */

    ArgNum++;
    if ( ArgV != NULL )
      *ArgV++ = StrPtr;

    for ( ; ; )
    {
      DoCopy = 1;
      Num = 0;

      while( (CH = *CPtr) == '\\' )
      {
        Num++; CPtr++;
        CH = *CPtr;
      }
      if ( CH == '\"' )
      {
        if ( (Num & 1) == 0 )
        {
          if ( Quoted )
          {
            if ( CPtr[1] == '\"' )
              CPtr++;
            else
              DoCopy = 0;
          }
          else
          {
            DoCopy = 0;
          }

          Quoted = !Quoted;
        }

        Num >>= 1;
      }

      while ( Num-- )
      {
        ChrNum++;
        if ( StrPtr != NULL )
          *StrPtr++ = '\\';
      }

      if ( (CH = *CPtr) == '\0' || (!Quoted && (CH == ' ' || CH == '\t')) )
        break;

      if ( DoCopy )
      {
        ChrNum++;
        if ( StrPtr != NULL )
          *StrPtr++ = CH;
      }

      CPtr++;
    }

    ChrNum++;
    if ( StrPtr != NULL )
      *StrPtr++ = '\0';

  }

  *NumArgs = ArgNum;
  *NumChar = ChrNum;
}

/* ------------------------------------------------------------------------- */

static CHAR** ParseCommandLine(INT* ArgC)
{
  CHAR** ArgV;
  CHAR   ModName[MAX_PATH];
  CHAR*  CmdLine;
  CHAR*  StrLine;
  UINT   NumArgs;
  UINT   NumChar;
  UINT   MemSize;

  /* get original command line */
  GetModuleFileName(NULL, ModName, MAX_PATH);
  if ( NULL == (CmdLine = GetCommandLine()) || CmdLine[0] == '\0' )
    CmdLine = ModName; /* use file path as command line */

  /* calc size of args */
  ExtractCmdLineParams(CmdLine, NULL, NULL, &NumArgs, &NumChar);

  /* alloc memory */
  MemSize = ((NumArgs + 1) * sizeof(CHAR*)) + NumChar;
  if ( NULL == (ArgV = (CHAR**)SmallAlloc(MemSize)) )
    return(NULL);

  /* reconstruct arguments */
  StrLine = (CHAR*)(ArgV) + ((NumArgs+1) * sizeof(CHAR*));
  ExtractCmdLineParams(CmdLine, ArgV, StrLine, &NumArgs, &NumChar);
  if ( NULL != ArgC ) *ArgC = NumArgs;

  return(ArgV); /* all ok */
}

/* ------------------------------------------------------------------------- */

/* unpacker work results */
#define ASPUNP_WORK_OK       0    /* all ok                              */
#define ASPUNP_WORK_ERROR   -1    /* some processing error               */
#define ASPUNP_WORK_PARAMS  -2    /* invalid paramethers in command line */


/* module entrypoint */
#ifdef _DEBUG
INT  CALLBACK WinMain(HINSTANCE hInst, HINSTANCE hPrevInst,
                      LPSTR CmdLine, INT CmdShow)
#else
VOID CALLBACK xWinMain()
#endif
{
  LPTOP_LEVEL_EXCEPTION_FILTER Filter;
  UNASP_CTX  Ctx;
  HANDLE  hThread;
  DWORD   ThrdID;
  UINT    uExitCode;
  CHAR**  ArgV;
  INT     ArgC;
  
  /* setup exception handler */
  Filter = (LPTOP_LEVEL_EXCEPTION_FILTER)(LastExceptionHandler);
  SetUnhandledExceptionFilter(Filter);
  
  /* parse command line */
  dg_memset(&Ctx, 0, sizeof(UNASP_CTX));
  if ( NULL == (ArgV = ParseCommandLine(&ArgC)) )
    ExitProcess(ASPUNP_WORK_ERROR);

  if ( ArgC == 1 )
  {
    /* dispatch gui */
    InitCommonControls();
    Ctx.hInst = GetModuleHandle(NULL);
    uExitCode = DialogBoxParam(Ctx.hInst, MAKEINTRESOURCE(IDD_MAINDLG), NULL,
      (DLGPROC)(DlgProc), (LPARAM)(&Ctx));
  }
  else
  {
    /* silent mode */
    if ( ArgC < 3 )
    {
      SmallFree(ArgV);
      ExitProcess(ASPUNP_WORK_PARAMS);
    }

    /* extract file names */
    Ctx.InpName = ArgV[1];
    Ctx.OutName = ArgV[2];

    /* dispatch working thread */
    if ( NULL == (hThread = CreateThread(NULL,0,UThrdProc,&Ctx,0,&ThrdID)) )
    {
      SmallFree(ArgV);
      ExitProcess(ASPUNP_WORK_ERROR);
    }

    /* wait til thread stop */
    uExitCode = (UINT)(ASPUNP_WORK_ERROR);
    while( GetExitCodeThread(hThread, &uExitCode) )
    {
      if ( uExitCode == STILL_ACTIVE )
      {
        uExitCode = (UINT)(ASPUNP_WORK_ERROR);
        Sleep(128);
        continue;
      }

      uExitCode = (uExitCode == 0) ? (ASPUNP_WORK_ERROR) : (ASPUNP_WORK_OK);
      break;
    }
  }

  SmallFree(ArgV);
  ExitProcess(uExitCode);

#ifdef _DEBUG
  return(0);
#endif
}

/* ------------------------------------------------------------------------- */
